#include "CHandleConnections.h"
#include "CShells.h"

CHandleConnections::CHandleConnections()
{
}

CHandleConnections::CHandleConnections(LOGINFO* _log_info)
{
	assert(_log_info);
	log = _log_info;

	shell = new (std::nothrow) CShells();
	assert(shell);

	users = new (std::nothrow) USERINFO[log->max_users];
	assert(users);
	for (unsigned int x = 0; x < log->max_users; x++)
		memset(&users[x], 0, sizeof(USERINFO));

	current_users = 0;
}

CHandleConnections::~CHandleConnections()
{
	if (shell)
		delete shell;

	if (users)
		delete users;

	log = 0;
	current_users = 0;
}

void CHandleConnections::HandleSockets(SOCKET l_socket, int* socket_list, fd_set* fd)
{
	assert(fd && socket_list);

	if (FD_ISSET(l_socket, fd))
		NewConnection(l_socket, socket_list);

	for (unsigned int n = 0; n < log->max_users; n++)
	{
		// if there is something to read on n (number)
		if (FD_ISSET(socket_list[n], fd))
			ReadFromSocket(n, socket_list);
	}
	return;
}

void CHandleConnections::NewConnection(SOCKET l_socket, int *socket_list)
{
	assert(socket_list);
	if (log->TimeoutExpired && log->UseLog)
	{
		std::cout << std::endl;
		log->TimeoutExpired = false;
	}
	sockaddr_in socket_info;
	int l_sockaddr = sizeof(socket_info);
	int user_sock = (int)accept(l_socket, (sockaddr *)&socket_info, &l_sockaddr);
	if (user_sock < 0)
	{
		if (log->UseLog)
			std::cout << "[-] accepting a socket on the listening socket (" <<
			l_socket << ") returned error " << errno << std::endl;

		return;
	}

	if (current_users >= log->max_users)
	{
		std::string text = "[*] Max users reached, try again later\n";
		send(user_sock, text.c_str(), (int)text.length(), 0);
		closesocket(user_sock);
		if (log->UseLog)
		{
			std::cout << "[*] Closing incoming connection from " << 
			inet_ntoa(socket_info.sin_addr) << " - max users reached" << std::endl;
		}
		return;
	}

	for (unsigned int id = 0; id < log->max_users; id++)
	{
		if (socket_list[id] == 0)
		{
			socket_list[id] = user_sock;
			AddConnectionToList(id, user_sock, socket_info);
			if (log->UseLog)
				std::cout << "[*] Incoming connection from " <<	inet_ntoa(socket_info.sin_addr);

			if (log->pass)
			{
				if (log->UseLog)
					std::cout << " - waiting for password" << std::endl;

				std::string req_pass = "[*] Server requires authentication, please identify yourself: ";
				send(user_sock, req_pass.c_str(), (int)req_pass.length(), 0);
			}
			else
			{
				if (log->UseLog)
				{
					std::cout << std::endl;
					std::cout << "[*] " << inet_ntoa(users[id].s_info.sin_addr) << 
					" has shell: " << shelltypes[log->ShellType].descr;
					if (log->ShellProgram)
					{
						std::cout << " (" << log->ShellProgram << ")";
						users[id].shell_program = log->ShellProgram;
					}
					std::cout << std::endl;
				}
				users[id].shell_t = shelltypes[log->ShellType].id;
				shell->StartShell(&users[id]);
			}
			break;
		}
	}

	if (current_users >= log->max_users && log->UseLog)
		std::cout << "[*] Max users reached (" << log->max_users << ")" << std::endl;
}

void CHandleConnections::ReadFromSocket(unsigned int id, int* socket_list)
{
	assert(socket_list);
	if (log->TimeoutExpired && log->UseLog)
	{
		std::cout << std::endl;
		log->TimeoutExpired = false;
	}
	memset(users[id].s_recv, 0, sizeof(users[id].s_recv));
	users[id].b_recv = recv(users[id].s_id, users[id].s_recv, sizeof(users[id].s_recv), 0);
	if (users[id].b_recv < 1)
	{
		if (log->UseLog)
			std::cout << "[*] " << inet_ntoa(users[id].s_info.sin_addr) << 
			" closed connection, closing shell and disconnecting client" << std::endl;

		if (users[id].shell_t == SHELL_DEFAULT_PROGRAM && users[id].shell.process.pi.hProcess)
			TerminateProcess(users[id].shell.process.pi.hProcess, 0);

		RemoveConnectionFromList(id, socket_list);
		return;
	}

	/*
		If the server has a password set AND the user is not yet authenticated,
		compare the given password from the server password set.
		If the passwords didnt match close and remove the connection. If they did match
		set the user as authed, and if -s = 3 send 'shell>'
	*/
	if ((log->pass) && (!users[id].is_authed))
	{
		if (!UserPasswordIsValid(&users[id]))
		{
			closesocket(users[id].s_id);
			RemoveConnectionFromList(id, socket_list);
			return;
		}
		users[id].is_authed = 1;
	}

	/*
		If the user doenst has a running shell already,
		and the server has a -s option (1-3), then start a shell and set the shell type
		at the user pointer. That means if there aint a -s specified or the user
		has already a running shell, it'll skip this. Default users[id].shell_t = 0.
	*/
	if (!users[id].shell_t && log->ShellType)
	{
		if (log->UseLog)
		{
			std::cout << "[*] " << inet_ntoa(users[id].s_info.sin_addr) << 
			" has shell: " << shelltypes[log->ShellType].descr;
			if (log->ShellProgram)
			{
				std::cout << " (" << log->ShellProgram << ")";
				users[id].shell_program = log->ShellProgram;
			}
			std::cout << std::endl;
		}
		users[id].shell_t = shelltypes[log->ShellType].id;
		shell->StartShell(&users[id]);
		return;
	}
	if (log->UseLog)
	{
		std::cout << "[*] " << inet_ntoa(users[id].s_info.sin_addr) << 
		" sended command: " << users[id].s_recv;
	}
	int err = shell->WriteToShell(&users[id]);
	switch(err)
	{
		case WRITE_SHELL_FAILED:
		{
			if (log->UseLog)
			{
				std::cout << "[-] Failed to write to the shell of " <<
				inet_ntoa(users[id].s_info.sin_addr) << " - closing connection" << std::endl;
			}
			std::string text = "\n[-] An error has occured on the server side, closing connection\n";
			send(users[id].s_id, text.c_str(), (int)text.length(), 0);
			RemoveConnectionFromList(id, socket_list);
			break;
		}
		case WRITE_SHELL_EXIT:
		{
			if (log->UseLog)
			{
				std::cout << "[*] " << inet_ntoa(users[id].s_info.sin_addr) << 
				" closed shell - closing connection" << std::endl;
			}
			RemoveConnectionFromList(id, socket_list);
			break;
		}
		case WRITE_SHELL_OK:
		default:
			break;
	}
}

void CHandleConnections::AddConnectionToList(unsigned int id, SOCKET socket_id, sockaddr_in socket_info)
{
	memset(&users[id], 0, sizeof(USERINFO));
	users[id].s_id = socket_id;
	users[id].s_info = socket_info;
	users[id].user_id = current_users;
	current_users++;
}

void CHandleConnections::RemoveConnectionFromList(unsigned int id, int* socket_list)
{
	closesocket(socket_list[id]);
	socket_list[id] = 0;
	memset(&users[id], 0, sizeof(USERINFO));
	current_users--;
}

bool CHandleConnections::UserPasswordIsValid(USERINFO* user)
{
	user->s_recv[strlen(user->s_recv)-1] = 0;
	if (_stricmp(log->pass, user->s_recv) != 0)
	{
		if (log->UseLog)
		{
			std::cout << "[*] Invalid password from " <<
			inet_ntoa(user->s_info.sin_addr) << " - closing connection" << std::endl;
		}
		return false;
	}
	if (log->UseLog)
	{
		std::cout << "[*] " << inet_ntoa(user->s_info.sin_addr) << 
		" authorized: access granted, current users: " << current_users << std::endl;
	}
	return true;
}

void CHandleConnections::KeepAlive(int *socket_list, fd_set* fd, SOCKET l_socket)
{
	std::string idletext = ".";
	for (unsigned int x = 0; x < log->max_users; x++)
	{
		if (socket_list[x] == l_socket)
			continue;
		else if (socket_list[x] != 0)
			send(socket_list[x], idletext.c_str(), (int)idletext.length(), 0);
	}
}

void CHandleConnections::CloseAllSockets(int* socketlist)
{
	std::string txt = "\n[*] Server going down, closing connection and shell\n";
	for (unsigned int n = 0; n < log->max_users; n++)
	{
		if (socketlist[n] != 0)
		{
			send(socketlist[n], txt.c_str(), (int)txt.length(), 0);
			shell->CloseShell(&users[n]);
		}
	}
}